/*
 * Copyright (c) 1998,1999 Trans New Technology, Inc. and
 *                         Hiroki NAKANO <cas@trans-nt.com>
 * All rights reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that this notice is retained,
 * the conditions in the following notices are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *	Trans New Technology, Inc. (in Japan)
 *
 *
 * THIS SOFTWARE IS PROVIDED BY TRANS NEW TECHNOLOGY, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL TRANS NEW TECHNOLOGY, INC. BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * ICOM UX-136 Wireless LAN Card
 *   (AMD Am79C930 Firmware ver.2.0-Japan)
 *
 *   machine-independent module
 *
 */

#include "ux.h"
#define	STATIC	/* for debug */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/device.h>
#include <sys/proc.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/if_dl.h>
#include <net/netisr.h>
#include <net/route.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#ifdef INET6
#ifndef INET
#include <netinet/in.h>
#endif
#include <netinet6/in6_ifattach.h>
#endif

#include "bpfilter.h"
#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif /* NBPFILTER > 0 */

#include <machine/cpu.h>
#include <machine/param.h>
#include <machine/cpufunc.h>
#include <machine/clock.h>
#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>

#if defined(__FreeBSD__)
#include <i386/isa/icu.h>
#include <i386/isa/isa.h>
#include <i386/isa/isa_device.h>
#include "card.h"
#include <sys/select.h>
#if defined(UX_FREEBSD226)
#include <pccard/card.h>		/* FreeBSD2.2.6 */
#else
#include <pccard/cardinfo.h>		/* FreeBSD2.2.7 */
#endif
#include <pccard/driver.h>
#include <pccard/slot.h>
#endif /* __FreeBSD__ */

#include <machine/stdarg.h>

#if defined(__FreeBSD__)
#include <i386/isa/if_uxtypes.h>
#include <dev/ux/am79c930reg.h>
#include <dev/ux/am79c930var.h>
#include <dev/ux/am79c930ioctl.h>
#include <i386/isa/if_ux.h>
#endif /* __FreeBSD__ */

/*
 *  Macro for Messages to Console.
 */
#define	DN(string)	"%s: " string , sc->ux_devname
#if 1
#define	DP(args)	printf args
/* DP((DN("messages to console.\n"), some parameters)) */
#define DP9(args)
#else
#define	DP(args)
#define DP9(args)
#endif

/* Prototype Declaration of Static Functions */
/*    produced automatically by 'make proto' */
STATIC int ux_issuecommand(struct ux_softc *sc, int cmd);
STATIC int ux_waitresult(struct ux_softc *sc, u_int8_t *statusp);
STATIC int ux_execcommand(struct ux_softc *sc, int cmd, u_int8_t *statusp);
STATIC int ux_beginintr(struct ux_softc *sc);
STATIC void ux_enableintr(struct ux_softc *sc, int sw);
STATIC void ux_clearintr(struct ux_softc *sc);
STATIC int ux_setintrmask(struct ux_softc *sc, int mask);
STATIC int ux_lockintr(struct ux_softc *sc);
STATIC void ux_unlockintr(struct ux_softc *sc);
STATIC int ux_checkfirmware(struct ux_softc *sc);
STATIC int ux_checkselfteststatus(struct ux_softc *sc, int *statusp, char **msgp);
STATIC int ux_getmib(struct ux_softc *sc, void *ptr, int mibtype, int start, int size);
STATIC int ux_setmib(struct ux_softc *sc, void *ptr, int mibtype, int start, int size);
STATIC int ux_getmacaddr(struct ux_softc *sc, ux_uint48_t *addr);
STATIC int ux_setpromiscuous(struct ux_softc *sc, int sw);
STATIC void ux_inittxbuf(struct ux_softc *sc, u_int32_t top, u_int32_t size);
STATIC int ux_alloctxbuf(struct ux_softc *sc, u_int32_t *topp, u_int32_t *sizep);
STATIC void ux_freetxbuf(struct ux_softc *sc, u_int32_t top, u_int32_t size);
STATIC int ux_inittx(struct ux_softc *sc);
STATIC int ux_appendtxqueue(struct ux_softc *sc, u_int32_t *tailp, struct mbuf *m);
STATIC int ux_mbuftocard(struct ux_softc *sc, struct mbuf *m0, struct Am79C930_Tx_Desc *tdp, u_int32_t *tdponcard);
STATIC void ux_linkdesc(struct ux_softc *sc, u_int32_t from, u_int32_t to);
STATIC void ux_cleanuptx(struct ux_softc *sc);
STATIC int ux_initrx(struct ux_softc *sc);
STATIC int ux_scanrxqueue(struct ux_softc *sc, u_int32_t *descp);
STATIC int ux_input(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp);
STATIC int ux_cardtomem(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp, unsigned int offset, unsigned int size, void *buf);
STATIC int ux_cardtombuf(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp, unsigned int offset, struct mbuf **mp);
STATIC int ux_beginscan(struct ux_softc *sc, struct Am79C930_ESS_ID *essid);
STATIC int ux_recvproberesp(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp);
STATIC int ux_endscan(struct ux_softc *sc);
STATIC int ux_startbss(struct ux_softc *sc);
STATIC int ux_joinbss(struct ux_softc *sc, struct ux_bssinfo *bss);
STATIC int ux_initmibs(struct ux_softc *sc);
STATIC int ux_inittxrx(struct ux_softc *sc);
STATIC int ux_prepend80211header(struct ux_softc *sc, struct mbuf **mp);
STATIC void ux_remove80211header(struct ux_softc *sc, struct mbuf **mp);
STATIC void ux_resetstart(struct ux_softc *sc);
STATIC void ux_resettimo(void *arg);
/* END of Prototype Declaration of Static Functions */

/*
 *  issue command and wait for response
 */

STATIC int
ux_issuecommand(struct ux_softc *sc, int cmd)
{
	/* check if it's ready. */
	if(UX_MEM_READ1(sc,
			UX_OFFSETOF(Am79C930_Command_Block, Command,
				AM79C930_ADDR_COMMAND_BLOCK)) != 0) {
		if(UX_MEM_READ1(sc,
			UX_OFFSETOF(Am79C930_Command_Block, Command_Status,
				AM79C930_ADDR_COMMAND_BLOCK)) == 0) {
			printf("%s: ux_issuecommand(): device busy"
				" -- executing command.\n", sc->ux_devname);
			return EBUSY;
		}
		UX_MEM_WRITE1(sc, 0,
			UX_OFFSETOF(Am79C930_Command_Block, Command,
				AM79C930_ADDR_COMMAND_BLOCK));
	}

	UX_MEM_WRITE1(sc, 0,
		UX_OFFSETOF(Am79C930_Command_Block, Command_Status,
			AM79C930_ADDR_COMMAND_BLOCK));

	DP9((DN("issue command. cmd=%d\n"), cmd));

	/* issue command */
	UX_MEM_WRITE1(sc, cmd,
		UX_OFFSETOF(Am79C930_Command_Block, Command,
			AM79C930_ADDR_COMMAND_BLOCK));

	return 0;
}

STATIC int
ux_waitresult(struct ux_softc *sc, u_int8_t *statusp)
{
	int cmd, status, counter;

	/* has a command been issued? */
	cmd = UX_MEM_READ1(sc, UX_OFFSETOF(Am79C930_Command_Block, Command,
						AM79C930_ADDR_COMMAND_BLOCK));
	if(cmd == 0) {
		printf("%s: ux_waitresult(): no command issued.\n",
							sc->ux_devname);
		return EIO;
	}

	/* wait for complete command. give up in 10ms */
	for(counter = 0; counter < 100; counter++) {
		status = UX_MEM_READ1(sc,
			UX_OFFSETOF(Am79C930_Command_Block, Command_Status,
				AM79C930_ADDR_COMMAND_BLOCK));
		if(status != 0) break;
		DELAY(100);	/* 100us */
	}

	if(status == 0) {
		printf("%s: ux_waitresult(): timeout.(10ms)\n", sc->ux_devname);
		return EIO;
	}

	UX_MEM_WRITE1(sc, 0, UX_OFFSETOF(Am79C930_Command_Block, Command,
						AM79C930_ADDR_COMMAND_BLOCK));
	UX_MEM_WRITE1(sc, 0, UX_OFFSETOF(Am79C930_Command_Block, Command_Status,
						AM79C930_ADDR_COMMAND_BLOCK));

	DP9((DN("finished command. cmd=%d status=%d time=%dus\n"),
					cmd, status, counter * 100));

	if(statusp) *statusp = status;
	return status == AM79C930_CMDSTAT_COMPLETED ? 0 : EIO;
}

STATIC int
ux_execcommand(struct ux_softc *sc, int cmd, u_int8_t *statusp)
{
	int s = splimp();
	int error;

	error = ux_issuecommand(sc, cmd);
	if(error == 0) error = ux_waitresult(sc, statusp);
	splx(s);
	return error;
}

/*
 *  operating interrupt status/mask
 */

STATIC int
ux_beginintr(struct ux_softc *sc)
{
	u_int32_t status, status2;

	if(ux_lockintr(sc) != 0) {
		ux_clearintr(sc);
		DP((DN("ux_beginintr(): fail to lock. possibly lost intr.\n")));
		return 0;
	}

	status = UX_MEM_READ1(sc,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Interrupt_Status,
					AM79C930_ADDR_CONTROL_STATUS_BLOCK));
	status2 = UX_MEM_READ1(sc,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Interrupt_Status2,
					AM79C930_ADDR_CONTROL_STATUS_BLOCK));
	status = (status & 0xff) | ((status2 & 0xff) << AM79C930_INTR2_SHIFT);

	UX_MEM_WRITE1(sc, 0,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Interrupt_Status,
					AM79C930_ADDR_CONTROL_STATUS_BLOCK));
	UX_MEM_WRITE1(sc, 0,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Interrupt_Status2,
					AM79C930_ADDR_CONTROL_STATUS_BLOCK));
	
	ux_clearintr(sc);

	ux_unlockintr(sc);
	return status;
}

STATIC void
ux_enableintr(struct ux_softc *sc, int sw)
{
	u_int8_t gcr;
	int s = splimp();

	gcr = UX_IO_READ1(sc, AM79C930_IOREG_GCR);
	if(sw) gcr |= AM79C930_IOREG_GCR_ENECINT;
	else gcr &= ~AM79C930_IOREG_GCR_ENECINT;
	UX_IO_WRITE1(sc, gcr, AM79C930_IOREG_GCR);

	splx(s);
	return;
}

STATIC void
ux_clearintr(struct ux_softc *sc)
{
	u_int8_t gcr;
	int s = splimp();

	gcr = UX_IO_READ1(sc, AM79C930_IOREG_GCR);
	gcr |= AM79C930_IOREG_GCR_ECINT;
	UX_IO_WRITE1(sc, gcr, AM79C930_IOREG_GCR);

	splx(s);
	return;
}

STATIC int
ux_setintrmask(struct ux_softc *sc, int mask)
{
	int error;

	error = ux_lockintr(sc);
	if(error != 0) return error;

	UX_MEM_WRITE1(sc, mask & 0xff,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Interrupt_Mask,
				AM79C930_ADDR_CONTROL_STATUS_BLOCK));
	UX_MEM_WRITE1(sc, (mask >> AM79C930_INTR2_SHIFT) & 0xff,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Interrupt_Mask2,
				AM79C930_ADDR_CONTROL_STATUS_BLOCK));

	ux_unlockintr(sc);
	return 0;
}

STATIC int
ux_lockintr(struct ux_softc *sc)
{
	u_int8_t lh;
	int counter;

	for(counter = 0; counter < 100; counter++) {
		lh = UX_MEM_READ1(sc,
			UX_OFFSETOF(Am79C930_Control_Status_Block,
				Lockout_Host,
				AM79C930_ADDR_CONTROL_STATUS_BLOCK));
		if(lh == 0) {
			UX_MEM_WRITE1(sc, 1,
				UX_OFFSETOF(Am79C930_Control_Status_Block,
					Lockout_PCnet_MOBILE,
					AM79C930_ADDR_CONTROL_STATUS_BLOCK));
			
			lh = UX_MEM_READ1(sc,
				UX_OFFSETOF(Am79C930_Control_Status_Block,
					Lockout_Host,
					AM79C930_ADDR_CONTROL_STATUS_BLOCK));
			if(lh == 0) return 0;

			UX_MEM_WRITE1(sc, 0,
				UX_OFFSETOF(Am79C930_Control_Status_Block,
					Lockout_PCnet_MOBILE,
					AM79C930_ADDR_CONTROL_STATUS_BLOCK));
		}
		DELAY(100);	/* 100us */
	}

	/* timeout */
	return EBUSY;
}

STATIC void
ux_unlockintr(struct ux_softc *sc)
{
	UX_MEM_WRITE1(sc, 0,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Lockout_PCnet_MOBILE,
			AM79C930_ADDR_CONTROL_STATUS_BLOCK));
	return;
}

/*
 *  check if the firmware is ok.
 */

STATIC int
ux_checkfirmware(struct ux_softc *sc)
{
	struct Am79C930_Banner banner;
	int error, selftest;
	char *msg;
	u_int8_t status;

	/* check banner */

	UX_MEM_READM(sc, &banner, sizeof(banner), AM79C930_ADDR_BANNER);

	if(banner.Banner[sizeof(banner.Banner) - 1] != '\0') {
		int i;
		printf("%s: Abnormal Banner. [", sc->ux_devname);
		for(i = 0; i < sizeof(banner.Banner); i++) {
			unsigned char c = banner.Banner[i];
			if(c < 0x20 || c > 0x7e) printf("\\x%02x", c);
			else printf("%c", c);
		}
		printf("]\n");
		return EIO;
	}

	printf("%s: %s\n", sc->ux_devname, banner.Banner);

	/* check Self Test Status */
	error = ux_checkselfteststatus(sc, &selftest, &msg);
	DP((DN("SelfTest:%02x %s\n"), selftest, msg));
	if(error != 0) {
		printf(DN("Self Test Failure(%02x) %s\n"), selftest, msg);
		return EIO;
	}

	/* try a test command */
	error = ux_execcommand(sc, AM79C930_CMD_TEST_INTERFACE_NOP, &status);
	if(error != 0) {
		printf(DN("fail a test command. status=%d\n"), status);
		return EIO;
	}
	
	/* all ok. */
	return 0;
}

STATIC int
ux_checkselfteststatus(struct ux_softc *sc, int *statusp, char **msgp)
{
	static struct { unsigned char status, retval, *msg; } ststab[] = {
		{    0, EBUSY, "Initial State -- Self Test just started." },
		{    1, EBUSY, "PCnet-MOBILE firmware has started its firmware checksum test." },
		{    2, EBUSY, "PCnet-MOBILE firmware has started its hardware tests." },
		{    3, EBUSY, "PCnet-MOBILE firmware is initializing the MIB." },
		{ 0xff, 0, "PCnet-MOBILE firmware has completed all of its test successfully." },
		{ 0xfa, EIO, "MIB checksum failure" },
		{ 0xfb, EIO, "Radio failure" },
		{ 0xfc, EIO, "PCnet-MOBILE failure" },
		{ 0xfd, EIO, "Flash PROM checksum failure" },
		{ 0xfe, EIO, "RAM failure" },
		/* default for status except for above */
		{    0, EIO, "Unknown status" },
	};

	int status, i;

	status = UX_MEM_READ1(sc,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Self_Test_Status,
			AM79C930_ADDR_CONTROL_STATUS_BLOCK));

	for(i = 0; i < sizeof(ststab) / sizeof(ststab[0]) - 1; i++) {
		if(ststab[i].status == status) break;
	}

	if(statusp) *statusp = status;
	if(msgp) *msgp = ststab[i].msg;
	return ststab[i].retval;
}

/*
 *  basic MIB operation
 */
#define	UX_GETMIB(sc,ptr,mibtype,mibstruct,from,to)	\
		ux_getmib(sc,ptr,mibtype,		\
			UX_OFFSETOF(mibstruct,from,0),	\
			UX_OFFSETOF(mibstruct,to,0)	\
			- UX_OFFSETOF(mibstruct,from,0)	\
			+ UX_SIZEOF(mibstruct,to))
#define	UX_SETMIB(sc,ptr,mibtype,mibstruct,from,to)	\
		ux_setmib(sc,ptr,mibtype,		\
			UX_OFFSETOF(mibstruct,from,0),	\
			UX_OFFSETOF(mibstruct,to,0)	\
			- UX_OFFSETOF(mibstruct,from,0)	\
			+ UX_SIZEOF(mibstruct,to))
STATIC int
ux_getmib(struct ux_softc *sc, void *ptr, int mibtype, int start, int size)
{
	struct Am79C930_CmdParam_Get_MIB_Variables cp;
	int s, error;

	if(size < 0 || size > sizeof(cp.Data)) {
		printf(DN("ux_getmib(): invalid size %d\n"), size);
		return EINVAL;
	}

	cp.Type = mibtype;
	cp.Size = size;
	cp.Index = start;

	s = splimp();

	UX_MEM_WRITEM(sc, &cp,
		UX_OFFSETOF(Am79C930_CmdParam_Get_MIB_Variables, Data, 0),
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK));

	error = ux_execcommand(sc, AM79C930_CMD_GET_MIB_VARIABLES, 0);

	if(error == 0) { /* When the command was finished successfully, */
		UX_MEM_READM(sc, ((u_int8_t *)ptr) + start, size,
			UX_OFFSETOF(Am79C930_CmdParam_Get_MIB_Variables, Data,
			UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
					AM79C930_ADDR_COMMAND_BLOCK)));
	}
	
	splx(s);
	return error;
}

STATIC int
ux_setmib(struct ux_softc *sc, void *ptr, int mibtype, int start, int size)
{
	struct Am79C930_CmdParam_Set_MIB_Variables cp;
	int s, error;

	if(size < 0 || size > sizeof(cp.Data)) {
		printf(DN("ux_setmib(): invalid size %d\n"), size);
		return EINVAL;
	}

	cp.Type = mibtype;
	cp.Size = size;
	cp.Index = start;

	s = splimp();

	UX_MEM_WRITEM(sc, &cp,
		UX_OFFSETOF(Am79C930_CmdParam_Set_MIB_Variables, Data, 0),
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK));
	UX_MEM_WRITEM(sc, ((u_int8_t *)ptr) + start, size,
		UX_OFFSETOF(Am79C930_CmdParam_Set_MIB_Variables, Data,
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK)));

	error = ux_execcommand(sc, AM79C930_CMD_SET_MIB_VARIABLES, 0);

	splx(s);
	return error;
}

/*
 *  get/set various information.
 */

STATIC int
ux_getmacaddr(struct ux_softc *sc, ux_uint48_t *addr)
{
	struct Am79C930_MAC_Address_Status_grp masg;
	int error;
	
	error = UX_GETMIB(sc, &masg, AM79C930_MIB_MAC_ADDRESS_STATUS_GRP,
		Am79C930_MAC_Address_Status_grp, aMAC_Address, aMAC_Address);

	if(error == 0) bcopy(&masg.aMAC_Address, addr, sizeof(*addr));

	return error;
}

STATIC int
ux_setpromiscuous(struct ux_softc *sc, int sw)
{
	struct Am79C930_MAC_MIB mm;
	int error;

	mm.aPromiscuous_Enable = sw ? 1 : 0;
	return UX_SETMIB(sc, &mm, AM79C930_MIB_MAC_MIB,
		Am79C930_MAC_MIB, aPromiscuous_Enable, aPromiscuous_Enable);
}

/*
 *  tx buffer management
 */

#ifndef UX_TXBUFBLOCKSIZE
#define UX_TXBUFBLOCKSIZE	0x80
#endif /* UX_TXBUFBLOCKSIZE */

STATIC void
ux_inittxbuf(struct ux_softc *sc, u_int32_t top, u_int32_t size)
{
	sc->ux_txbufoffset = top;
	sc->ux_txbufsize = size;
	bzero(sc->ux_txbuffreemap, sizeof(sc->ux_txbuffreemap));
	return;
}

STATIC int
ux_alloctxbuf(struct ux_softc *sc, u_int32_t *topp, u_int32_t *sizep)
{
	u_int32_t size = *sizep;
	u_int32_t cbtop, cbno, cbsize;
	u_int32_t blk, blkno;
	int error;

	DP9((DN("ux_alloctxbuf(): size=%d\n"), size));

	if(size == 0) return EINVAL;

	cbsize = 0;
	blkno = 0;
	blk = sc->ux_txbufoffset;
	error = ENOBUFS;
	while(blk < sc->ux_txbufoffset + sc->ux_txbufsize) {
		unsigned int byte, bit;
		byte = blkno >> 3;
		bit = 0x80U >> (blkno & 7);

		if(byte >= sizeof(sc->ux_txbuffreemap)) break;

		if(! (sc->ux_txbuffreemap[byte] & bit)) { /* free block */
			if(cbsize == 0) {
				cbtop = blk;
				cbno = blkno;
			}
			cbsize += UX_TXBUFBLOCKSIZE;

			if(cbsize >= size) { /* satisfied! */
				sc->ux_txbuffreemap[byte] |= bit;
				while(cbno < blkno) {
					blkno--;
					sc->ux_txbuffreemap[blkno >> 3]
						|= 0x80U >> (blkno & 7);
				}
				*topp = cbtop;
				*sizep = cbsize;
				error = 0;			/* OK */
				break;
			}
		} else {	/* block in use */
			cbsize = 0;
		}

		blkno++;
		blk += UX_TXBUFBLOCKSIZE;
	}

#if 0
	/* print status */
	DP((DN("ux_alloctxbuf(): ")));
	blkno = 0;
	blk = sc->ux_txbufoffset;
	while(blk < sc->ux_txbufoffset + sc->ux_txbufsize) {
		unsigned int byte, bit;
		byte = blkno >> 3;
		bit = 0x80U >> (blkno & 7);

		if(byte >= sizeof(sc->ux_txbuffreemap)) break;

		DP((sc->ux_txbuffreemap[byte] & bit ? "x" : "."));

		blkno++;
		blk += UX_TXBUFBLOCKSIZE;
	}
	DP(("\n"));
#endif
	
	return error;
}

STATIC void
ux_freetxbuf(struct ux_softc *sc, u_int32_t top, u_int32_t size)
{
	u_int32_t blkno = (top - sc->ux_txbufoffset) / UX_TXBUFBLOCKSIZE;
	
	for(;; blkno++) {
		if(blkno >= sizeof(sc->ux_txbuffreemap)) break;	/* XXX */
		sc->ux_txbuffreemap[blkno >> 3] &= ~(0x80U >> (blkno & 7));
		if(size <= UX_TXBUFBLOCKSIZE) break;
		size -= UX_TXBUFBLOCKSIZE;
	}
	return;
}

/*
 *  tx
 */

STATIC int
ux_inittx(struct ux_softc *sc)
{
	struct Am79C930_Tx_Desc txd;
	u_int32_t dataq, dataqsize;
	struct Am79C930_CmdParam_Init_Tx cpit;
	int error = 0;
	int s;

	/* construct the first descriptor */
	bzero(&txd, sizeof(txd));
	txd.Next_Descriptor = 0x80000000;
	txd.Tx_State = 0;	/* owned by host */
	
	dataqsize = sizeof(txd);
	error = ux_alloctxbuf(sc, &dataq, &dataqsize);
	if(error != 0) return error;
	UX_MEM_WRITEM(sc, &txd, sizeof(txd), dataq);

	/* Init_Tx command */
	bzero(&cpit, sizeof(cpit));
	cpit.Data_Desc = dataq;
	DP9((DN("Tx Desc: data=0x%04x\n"), dataq));
	
	s = splimp();
	UX_MEM_WRITEM(sc, &cpit, sizeof(cpit),
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK));
	error = ux_execcommand(sc, AM79C930_CMD_INIT_TX, 0);
	sc->ux_txdeschead = dataq;	/* from which we clean up descs */
	sc->ux_txdesctail = dataq;	/* to which packets are appended */
	splx(s);
	
	return error;
}

STATIC int
ux_appendtxqueue(struct ux_softc *sc, u_int32_t *tailp, struct mbuf *m)
{
	struct Am79C930_Tx_Desc td;
	u_int32_t newdesc;
	u_int8_t state;
	int error, usetail;

	DP9((DN("ux_appendtxqueue(): len=%d\n"), m->m_pkthdr.len));

	bzero(&td, sizeof(td));
	td.Tx_State = 0;			/* first, owned by host */
	td.Next_Descriptor = 0x80000000U;
	td.Tx_Rate = 20;			/* 2Mbps */

	state = UX_MEM_READ1(sc,
			UX_OFFSETOF(Am79C930_Tx_Desc, Tx_State, *tailp));
	usetail = ! (state & (AM79C930_TX_DESC_STATE_OWN
			    | AM79C930_TX_DESC_STATE_DONE));

#if 0
	if(usetail) {
		DP((DN("ux_appendtxqueue(): use tail.\n")));
	}
#endif

	newdesc = usetail ? *tailp : 0;
	error = ux_mbuftocard(sc, m, &td, &newdesc);
	if(error != 0) return error;

	if(! usetail) {
		ux_linkdesc(sc, *tailp, newdesc);
		*tailp = newdesc;
	}

	DP9((DN("ux_appendtxqueue(): tail=0x%04x\n"), *tailp));

	/* at last, change owner to firmware */
	td.Tx_State |= AM79C930_TX_DESC_STATE_OWN;
	UX_MEM_WRITE1(sc, td.Tx_State,
			UX_OFFSETOF(Am79C930_Tx_Desc, Tx_State, newdesc));
	
	return 0;	/* ok */
}

STATIC int
ux_mbuftocard(struct ux_softc *sc, struct mbuf *m0, struct Am79C930_Tx_Desc *tdp, u_int32_t *tdponcard)
{
	struct mbuf *m;
	u_int32_t tdpc, tdpsz;
	u_int32_t tframe, tframesz;
	int error;

	if((m0->m_flags & M_PKTHDR) == 0) return EINVAL;

	if(*tdponcard == 0) {
		tdpsz = sizeof(struct Am79C930_Tx_Desc);
		error = ux_alloctxbuf(sc, &tdpc, &tdpsz);
		if(error != 0) return error;
	} else {
		tdpc = *tdponcard;
		tdpsz = sizeof(struct Am79C930_Tx_Desc);
	}

	if(m0->m_pkthdr.len <= tdpsz - sizeof(struct Am79C930_Tx_Desc)) {
		tframe = tdpc + sizeof(struct Am79C930_Tx_Desc);
	} else {
		tframesz = m0->m_pkthdr.len;
		error = ux_alloctxbuf(sc, &tframe, &tframesz);
		if(error != 0) {
			if(*tdponcard == 0) ux_freetxbuf(sc, tdpc, tdpsz);
			return error;
		}
	}

	tdp->Tx_Start_of_Frame = tframe;
	tdp->Tx_Length = m0->m_pkthdr.len;

	tframesz = m0->m_pkthdr.len;
	for(m = m0; m && tframesz > 0; m = m->m_next) {
		u_int32_t tsize = m->m_len < tframesz ? m->m_len : tframesz;
		UX_MEM_WRITEM(sc, m->m_data, tsize, tframe);
		tframe += tsize;
		tframesz -= tsize;
	}

	if(m || tframesz > 0) {
		printf(DN("mismatch about length, m_pkthdr and real chain.\n"));
	}
	
	UX_MEM_WRITEM(sc, tdp, sizeof(*tdp), tdpc);

	*tdponcard = tdpc;
	return 0;
}

STATIC void
ux_linkdesc(struct ux_softc *sc, u_int32_t from, u_int32_t to)
{
	UX_MEM_WRITE4(sc, (to | 0x80000000U),
		UX_OFFSETOF(Am79C930_Tx_Desc, Next_Descriptor, from));
	UX_MEM_WRITE4(sc, to,
		UX_OFFSETOF(Am79C930_Tx_Desc, Next_Descriptor, from));
	return;
}

STATIC void
ux_cleanuptx(struct ux_softc *sc)
{
 	u_int32_t tdp = sc->ux_txdeschead;
	struct Am79C930_Tx_Desc td;

	DP9((DN("ux_cleanuptx(): start. head=0x%04x\n"), tdp));
	
	for(;;) {
		UX_MEM_READM(sc, &td, sizeof(td), tdp);
		if(td.Next_Descriptor & 0x80000000U) break;

		DP9((DN("ux_cleanuptx(): tdp=0x%04x\n"), tdp));

		if((td.Tx_State & AM79C930_TX_DESC_STATE_OWN)) break;/*by fw*/
		if((td.Tx_State & AM79C930_TX_DESC_STATE_DONE) == 0) break;

		DP9((DN("ux_cleanuptx(): flag=0x%02x"
			" DIFS=A%d/F%d SIFS=A%d/F%d RTS=%d Data=%d\n"),
			td.Tx_State,
			td.Num_DIFS_Attempts, td.Num_DIFS_Failures,
			td.Num_SIFS_Attempts, td.Num_SIFS_Failures,
			td.Num_RTS_Attempts, td.Num_Data_Attempts));

		/* see with ux_mbuftocard() */
		if(td.Tx_Start_of_Frame
				== tdp + sizeof(struct Am79C930_Tx_Desc)) {
			ux_freetxbuf(sc, tdp,
				td.Tx_Length + sizeof(struct Am79C930_Tx_Desc));
		} else {
			ux_freetxbuf(sc, tdp, sizeof(struct Am79C930_Tx_Desc));
			ux_freetxbuf(sc, td.Tx_Start_of_Frame, td.Tx_Length);
		}
		tdp = td.Next_Descriptor & 0x7fff;
	}

	DP9((DN("ux_cleanuptx(): end. head=0x%04x\n"), tdp));

 	sc->ux_txdeschead = tdp;
	return;
} 

/*
 *  rx
 */

STATIC int
ux_initrx(struct ux_softc *sc)
{
	struct Am79C930_CmdParam_Enable_Receiver cper;
	int error;
	int s;

	s = splimp();
	error = ux_execcommand(sc, AM79C930_CMD_ENABLE_RECEIVER, 0);
	if(error != 0) {
		splx(s);
		return error;
	}
	UX_MEM_READM(sc, &cper, sizeof(cper),
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK));
	splx(s);
	
	sc->ux_rxdeschead = cper.Data_Desc;
	sc->ux_rxdeschead_mgmt = cper.PS_Poll_Mgmt_Desc;

	DP((DN("Rx Desc: data=0x%04x mgmt=0x%04x%s\n"),
		sc->ux_rxdeschead, sc->ux_rxdeschead_mgmt,
		(sc->ux_rxdeschead_mgmt & 0x80000000) ? "(disabled)" : ""));
	
	return 0;
}

STATIC int
ux_scanrxqueue(struct ux_softc *sc, u_int32_t *descp)
{
	struct Am79C930_Rx_Desc rd;
	u_int32_t desc, dnext;
	u_int8_t dstate, dstate_next;
	int count,last;

	dnext = *descp;
	dnext &= 0x80007fffU;

	/* it's a disabled queue if the descriptor is invalid. */
	if((dnext & 0x80000000U)
	|| dnext < sc->ux_rxbufoffset
	|| dnext >= sc->ux_rxbufoffset + sc->ux_rxbufsize) return 0;

	for(count = 0, last = 0; !last && count < 30; count++) {
		desc = dnext;

		desc &= 0x7fffU;
		UX_MEM_READM(sc, &rd, sizeof(rd), desc);
		rd.Rx_Start_of_Frame &= 0x7fffU;	/* firmware's bug X-( */

		dstate = rd.Host_Desc_State;

		if(dstate & AM79C930_RX_DESC_STATE_OWN) { /* abnormal */
			sc->ux_if.if_ierrors++;
			break;
		}

		if(! (dstate & AM79C930_RX_DESC_STATE_CONSUMED)) {
			if(dstate & AM79C930_RX_DESC_STATE_BUFFER_OFLO) {
				sc->ux_if.if_ierrors++;
				printf(DN("Rx Buffer Overflow.\n"));
			} else {
				sc->ux_if.if_ipackets++;
				ux_input(sc, &rd);
			}

			dstate |= AM79C930_RX_DESC_STATE_CONSUMED;
		}

		dnext = rd.Next_Descriptor;
		dnext &= 0x80007fffU;

		/* no more packet when the next descriptor is not valid */
		if((dnext & 0x80000000U)
		|| dnext < sc->ux_rxbufoffset
		|| dnext >= sc->ux_rxbufoffset + sc->ux_rxbufsize) last = 1;
		else {
			dnext &= 0x7fffU;
			dstate_next = UX_MEM_READ1(sc,
			UX_OFFSETOF(Am79C930_Rx_Desc, Host_Desc_State, dnext));
			if(dstate_next & AM79C930_RX_DESC_STATE_OWN) last = 1;
			else dstate |= AM79C930_RX_DESC_STATE_OWN;
		}
		
		UX_MEM_WRITE1(sc, dstate,
			UX_OFFSETOF(Am79C930_Rx_Desc, Host_Desc_State, desc));
	}

	if(! last) {
		DP((DN("count ended. rx disabled.\n")));
		ux_setintrmask(sc, AM79C930_INTR_RX);
	}
	*descp = desc;
	return 0;
}

STATIC int
ux_input(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp)
{
	u_int32_t pstart = rdp->Rx_Start_of_Frame & 0x7fffU;
	u_int16_t psize = rdp->Rx_Length;
	struct mbuf *m, m0;
	int error = 0;
	u_int16_t fc;
	struct ether_header eh;

	fc = UX_MEM_READ2(sc, pstart);
	switch(fc & AM79C930_80211HDR_FC_TYPE_MASK) {
	case AM79C930_80211HDR_FC_TYPE_DATA:
		error = ux_cardtomem(sc, rdp,
			sizeof(struct Am79C930_80211_Header), sizeof(eh), &eh);
		if(error != 0) return 0;	/* error X_X */

		error = ux_cardtombuf(sc, rdp,
			sizeof(struct Am79C930_80211_Header) + sizeof(eh), &m);
		if(error != 0) return 0;	/* error X_X */

		/* throw to bpf */
		m0.m_type = MT_HEADER;
		m0.m_flags = 0;
		m0.m_data = m0.m_dat;
		m0.m_nextpkt = 0;
		m0.m_next = m;
		m0.m_len = sizeof(eh);

		if(! sc->ux_rawmacbpf) {
			bcopy(&eh, m0.m_data, sizeof(eh));
		} else {
			m0.m_len += sizeof(struct Am79C930_80211_Header);
			error = ux_cardtomem(sc, rdp, 0, m0.m_len, m0.m_data);
		}
		if(error == 0) ux_tap(sc, &m0);
		error = 0;

		/* throw packet to the upper layer */
		if((sc->ux_if.if_flags & IFF_PROMISC) != 0) {
			if(bcmp(eh.ether_dhost,
				sc->ux_macaddr, sizeof(eh.ether_dhost)) == 0
			|| (eh.ether_dhost[0] & 1) != 0) {
				ether_input(&sc->ux_if, &eh,  m);
			} else {
				m_freem(m);
			}
		} else {
			ether_input(&sc->ux_if, &eh,  m);
		}
		break;

	case AM79C930_80211HDR_FC_TYPE_MANAGEMENT:
		switch(fc & AM79C930_80211HDR_FC_SUBTYPE_MASK) {
		case AM79C930_80211HDR_FC_SUBTYPE_PROBE_RESPONSE:
			error = ux_recvproberesp(sc, rdp);
			/* fall through... */
		}
		/* fall through... */
	default:
		if(sc->ux_rawmacbpf) {
			error = ux_cardtombuf(sc, rdp, 0, &m);
			if(error != 0) return 0;	/* error X_X */
			ux_tap(sc, m);
			m_freem(m);
		}
		break;
	}

	return error;
}

STATIC int
ux_cardtomem(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp, unsigned int offset, unsigned int size, void *buf)
{
	u_int32_t pdata, psize;

	pdata = rdp->Rx_Start_of_Frame;
	psize = rdp->Rx_Length;

	if(psize < offset) {
		DP((DN("ux_cardtomem(): offset over end of packet.\n")));
		return EINVAL;
	}
	pdata += offset;
	psize -= offset;
	if(psize < size) {
		DP((DN("ux_cardtomem(): too short packet.\n")));
		return EINVAL;
	}

	UX_MEM_READM(sc, buf, size, pdata);
	return 0;
}

STATIC int
ux_cardtombuf(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp, unsigned int offset, struct mbuf **mp)
{
	struct mbuf *m, *m0, *m1;
	u_int32_t pdata, psize, tsize, msize;
	u_int32_t align;

	DP9((DN("ux_cardtombuf() begins.\n")));

	pdata = rdp->Rx_Start_of_Frame;
	psize = rdp->Rx_Length;
	DP9((DN("pdata=0x%04x psize=%d\n"), pdata, psize));

	if(psize < offset) return EINVAL;
	pdata += offset;
	psize -= offset;

	/* get the first mbuf */
	MGETHDR(m0, M_DONTWAIT, MT_HEADER);
	if(m0 == 0) return ENOBUFS;
	msize = MHLEN;
	if(psize >= MINCLSIZE) {
		MCLGET(m0, M_DONTWAIT);
		if(m0->m_flags & M_EXT) msize = m0->m_ext.ext_size;
	}
	m0->m_pkthdr.len = psize;
	m0->m_pkthdr.rcvif = &sc->ux_if;

	/* alignment issue */
	align = (u_int32_t)m0->m_data & 3;
	align = (2 - align) & 3;
	m0->m_data += align;
	msize -= align;

	for(m = m0;;) {
		tsize = msize < psize ? msize : psize;

		DP9((DN("m=0x%08x pdata=0x%04x tsize=%d msize=%d psize=%d\n"),
			m, pdata, tsize, msize, psize));

		UX_MEM_READM(sc, m->m_data, tsize, pdata);
		m->m_len = tsize;
		pdata += tsize;
		psize -= tsize;

		if(psize <= 0) break;

		MGET(m1, M_DONTWAIT, MT_DATA);
		if(m1 == 0) break;
		msize = MLEN;

		if(psize >= MINCLSIZE) {
			MCLGET(m1, M_DONTWAIT);
			if(m1->m_flags & M_EXT) msize = m1->m_ext.ext_size;
		}

		m->m_next = m1;
		m = m1;
	}

	if(psize > 0) {
		m_freem(m0);
		return ENOBUFS;
	}

	*mp = m0;
	return 0;
}

/*
 *  scan
 */

STATIC int
ux_beginscan(struct ux_softc *sc, struct Am79C930_ESS_ID *essid)
{
	struct Am79C930_CmdParam_Scan cpsc;
	struct Am79C930_MAC_MIB mm;
	int error;
	int s;

	sc->ux_nowbss.uxbss_valid = 0;

	bcopy(essid, &mm.aDesired_ESS_ID, sizeof(mm.aDesired_ESS_ID));
	error = UX_SETMIB(sc, &mm, AM79C930_MIB_MAC_MIB,
		Am79C930_MAC_MIB, aDesired_ESS_ID, aDesired_ESS_ID);
	if(error != 0) return error;

	bzero(&cpsc, sizeof(cpsc));
	cpsc.Set_Channel = 14;
	cpsc.Suspend = 0;
	cpsc.Duration = 1000; /* msec */
	
	s = splimp();		
	UX_MEM_WRITEM(sc, &cpsc, sizeof(cpsc),
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK));
	error = ux_execcommand(sc, AM79C930_CMD_SCAN, 0);
	splx(s);
	
	return error;
}

STATIC int
ux_recvproberesp(struct ux_softc *sc, struct Am79C930_Rx_Desc *rdp)
{
	struct ux_bssinfo *bss = &sc->ux_nowbss;
	struct Am79C930_80211_Header wh;
	struct Am79C930_80211_BeaconHeader bcn;
	int error = 0;
	int pos = 0;

	if(sc->ux_nowbss.uxbss_valid) return 0;	/* the first one wins. */

	DP((DN("recv PROBERESP\n")));

	error = ux_cardtomem(sc, rdp, pos, sizeof(wh), &wh);
	if(error != 0) return EINVAL;
	pos += sizeof(wh);

	error = ux_cardtomem(sc, rdp, pos, sizeof(bcn), &bcn);
	if(error != 0) return EINVAL;
	pos += sizeof(bcn);

	DP((DN("capinfo=0x%04x\n"), bcn.capability_info));

	if((bcn.capability_info & 0x1f) != 2) return 0;

	while(pos < rdp->Rx_Length - 4) {	/* ignore 4 bytes of FCS */
		ux_uint8_t elm[256];

		if(pos + 2 > rdp->Rx_Length - 4) {
			DP((DN("broken probe response.\n")));
			return EINVAL;
		}
		error = ux_cardtomem(sc, rdp, pos, 2, elm);
		if(error != 0) break;
		pos += 2;

		if(pos + elm[1] > rdp->Rx_Length - 4) {
			DP((DN("broken probe response.\n")));
			return EINVAL;
		}
		error = ux_cardtomem(sc, rdp, pos, elm[1], &elm[2]);
		if(error != 0) break;
		pos += elm[1];

		switch(elm[0]) {
		case AM79C930_80211_ELMID_SSID: {
			int i;
			int idsize = sizeof(bss->uxbss_essid);
			if(idsize > elm[1] + 2) idsize = elm[1] + 2;
			bcopy(elm, bss->uxbss_essid, idsize);

			DP((DN("SSID: size=%d data=["), elm[1]));
			for(i = 2; i < idsize; i++) { DP(("%c", elm[i])); }
			DP(("]\n"));
			} break;
		}
	}

	bcopy(&wh.Address[2], &bss->uxbss_bssid, sizeof(bss->uxbss_bssid));
	bcopy(&bcn.timestamp, &bss->uxbss_timestamp,
					sizeof(bss->uxbss_timestamp));
	bss->uxbss_reftime = rdp->Local_Time;

	bss->uxbss_valid = 1;

	DP((DN("BSSID=%02x:%02x:%02x:%02x:%02x:%02x\n"),
			bss->uxbss_bssid[0], bss->uxbss_bssid[1],
			bss->uxbss_bssid[2], bss->uxbss_bssid[3],
			bss->uxbss_bssid[4], bss->uxbss_bssid[5]));

	return 0;
}

STATIC int
ux_endscan(struct ux_softc *sc)
{
	if(sc->ux_nowbss.uxbss_valid) {
		ux_joinbss(sc, &sc->ux_nowbss);
	} else {
		ux_startbss(sc);
	}
	return 0;
}

/*
 *  management
 */

STATIC int
ux_startbss(struct ux_softc *sc)
{
	struct Am79C930_CmdParam_Sync cpsy;
	struct Am79C930_MAC_MIB mm;
	struct Am79C930_MAC_Mgmt_MIB gm;
	int error;
	int s;

	DP((DN("start BSS.\n")));

	/* set ESS ID */
	bzero(&gm.aCurrent_ESS_ID, sizeof(gm.aCurrent_ESS_ID));

	error = UX_SETMIB(sc, &gm, AM79C930_MIB_MAC_MGMT_MIB,
		Am79C930_MAC_Mgmt_MIB, aCurrent_BSS_ID, aCurrent_ESS_ID);
	if(error != 0) return error;

	/* SYNC command */
	bzero(&cpsy, sizeof(cpsy));
	cpsy.Set_Channel = 14;
	cpsy.Start_BSS = 1;

	s = splimp();		
	UX_MEM_WRITEM(sc, &cpsy, sizeof(cpsy),
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK));
	error = ux_execcommand(sc, AM79C930_CMD_SYNC, 0);
	splx(s);
	
	error = UX_GETMIB(sc, &gm, AM79C930_MIB_MAC_MGMT_MIB,
		Am79C930_MAC_Mgmt_MIB, aCurrent_BSS_ID, aCurrent_ESS_ID);
	if(error != 0) return error;
	bcopy(&gm.aCurrent_BSS_ID, sc->ux_bssid, sizeof(sc->ux_bssid));

	return error;
}

STATIC int
ux_joinbss(struct ux_softc *sc, struct ux_bssinfo *bss)
{
	struct Am79C930_CmdParam_Sync cpsy;
	struct Am79C930_MAC_Mgmt_MIB gm;
	int error;
	int s;
	int i;

	DP((DN("join BSS.\n")));
	DP((DN("ESSID: type=%d size=%d data=["),
			bss->uxbss_essid[0], bss->uxbss_essid[1]));
	for(i = 2; i < bss->uxbss_essid[1] + 2; i++) {
		DP(("%c", bss->uxbss_essid[i]));
	}
	DP(("]\n"));
	DP((DN("BSSID=%02x:%02x:%02x:%02x:%02x:%02x\n"),
			bss->uxbss_bssid[0], bss->uxbss_bssid[1],
			bss->uxbss_bssid[2], bss->uxbss_bssid[3],
			bss->uxbss_bssid[4], bss->uxbss_bssid[5]));

	/* set BSS ID & ESS ID */
	bcopy(&bss->uxbss_bssid, &gm.aCurrent_BSS_ID,
					sizeof(gm.aCurrent_BSS_ID));
	bcopy(&bss->uxbss_essid, &gm.aCurrent_ESS_ID,
					sizeof(gm.aCurrent_ESS_ID));

	error = UX_SETMIB(sc, &gm, AM79C930_MIB_MAC_MGMT_MIB,
		Am79C930_MAC_Mgmt_MIB, aCurrent_BSS_ID, aCurrent_ESS_ID);
	if(error != 0) return error;

	/* SYNC command */
	bzero(&cpsy, sizeof(cpsy));
	cpsy.Set_Channel = 14;
	cpsy.Start_BSS = 0;
	cpsy.Ref_Time= bss->uxbss_reftime;
	bcopy(&bss->uxbss_timestamp, &cpsy.BSS_timestamp,
				sizeof(cpsy.BSS_timestamp));

	s = splimp();		
	UX_MEM_WRITEM(sc, &cpsy, sizeof(cpsy),
		UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
				AM79C930_ADDR_COMMAND_BLOCK));
	error = ux_execcommand(sc, AM79C930_CMD_SYNC, 0);
	splx(s);
	if(error != 0) return error;

	bcopy(bss->uxbss_bssid, sc->ux_bssid, sizeof(sc->ux_bssid));
	return 0;
}

STATIC int
ux_initmibs(struct ux_softc *sc)
{
	int error;
	struct Am79C930_Local_MIB lm;
	struct Am79C930_MAC_MIB mm;
	struct Am79C930_MAC_Mgmt_MIB gm;

	error = UX_GETMIB(sc, &lm, AM79C930_MIB_LOCAL_MIB,
		Am79C930_Local_MIB, Fragmentation_Dis, Fill_CFP);
	if(error != 0) return error;

	error = UX_GETMIB(sc, &mm, AM79C930_MIB_MAC_MIB,
		Am79C930_MAC_MIB, Reserved1, aDesired_ESS_ID);
	if(error != 0) return error;

	error = UX_GETMIB(sc, &gm, AM79C930_MIB_MAC_MGMT_MIB,
		Am79C930_MAC_Mgmt_MIB, aPower_Mgt_Mode, aCurrent_ESS_ID);
	if(error != 0) return error;

	/* Local MIB */
	lm.Fragmentation_Dis		= 0x01;
	lm.Add_PLCP_Dis			= 0x00;
	lm.MAC_Hdr_Prsv			= 0x00;
	lm.Rx_Mgmt_Que_En		= 0x00;
	lm.Re_Assembly_Dis		= 0x00;
	lm.Strip_PLCP_Dis		= 0x00;
	lm.Rx_Error_Dis			= 0x00;
	lm.Power_Saving_Mode_Dis	= 0x00;	/* ik = 1 */
	lm.Accept_All_Multicast_Dis	= 0x00;
	lm.Check_Seq_Cntl_Dis		= 0x00;
	lm.Flush_CFP_Queue_On_CF_End	= 0x00;
	lm.Network_Mode			= 0x00;
	lm.PWD_Lvl			= 0x00;
	lm.CFP_Mode			= 0x00;
	lm.Acting_as_AP			= 0x00;
	lm.Fill_CFP			= 0x00;
	
	/* MAC MIB */
	mm.aRTS_Threshold		= 0x08fd;
	mm.aCW_max			= 0x03ff;
	mm.aCW_min			= 0x001f;
	mm.aPromiscuous_Enable		= 0x00;
	mm.aShort_Retry_Limit		= 0x0f;
	mm.aLong_Retry_Limit		= 0x0f;
	mm.aMax_Frame_Length		= 0x0900;
	mm.aFragmentation_Threshold 	= 0x0900;
	mm.aProbe_Delay			= 0x287a;
	mm.aMin_Probe_Response_Time 	= 0x0005;
	mm.aMax_Probe_Response_Time 	= 0x0032;
	mm.aMax_Transmit_MSDU_Lifetime 	= 0x00008000;
	mm.aMax_Receive_MSDU_Lifetime 	= 0x00008000;
	mm.aStation_Basic_Rate		= 0x000a;

#if 0
	bzero(&mm.aDesired_ESS_ID, sizeof(mm.aDesired_ESS_ID));
#endif

	/* MAC Management MIB */
	gm.aPower_Mgt_Mode		= 0x00;
	gm.aScan_Mode			= 0x01;
	gm.aScan_State			= 0x00;
	gm.aDTIM_Period			= 0x05;
	gm.aATIM_Window			= 0x0000;
	gm.Wep_Required			= 0x00;
	gm.aBeacon_Period		= 0x0064;
	gm.aPassive_Scan_Duration	= 0x0064;
	gm.aListen_Interval		= 0x0000;
	gm.aMedium_Occupancy_Limit	= 0x01f4;
	gm.aMax_MPDU_Time		= 0x6537;
	gm.aCFP_Max_Duration		= 0x0033;
	gm.aCFP_Rate			= 0x05;
	gm.Do_Not_Receive_DTIMs 	= 0x00;
	gm.aStation_ID			= 0x0000;
	
#if 0
	bcopy(&sc->ux_bssid, &gm.aCurrent_BSS_ID, sizeof(gm.aCurrent_BSS_ID));
	bzero(&gm.aCurrent_ESS_ID, sizeof(gm.aCurrent_ESS_ID));
#endif

	/*------------*/

	error = UX_SETMIB(sc, &lm, AM79C930_MIB_LOCAL_MIB,
		Am79C930_Local_MIB, Fragmentation_Dis, Fill_CFP);
	if(error != 0) return error;

	error = UX_SETMIB(sc, &mm, AM79C930_MIB_MAC_MIB,
		Am79C930_MAC_MIB, Reserved1, aDesired_ESS_ID);
	if(error != 0) return error;

	error = UX_SETMIB(sc, &gm, AM79C930_MIB_MAC_MGMT_MIB,
		Am79C930_MAC_Mgmt_MIB, aPower_Mgt_Mode, aCurrent_ESS_ID);
	if(error != 0) return error;

	return 0;
}

STATIC int
ux_inittxrx(struct ux_softc *sc)
{
	struct Am79C930_Local_MIB lm;
	int error;

	/* get buffer address from Local MIB */
	error = UX_GETMIB(sc, &lm, AM79C930_MIB_LOCAL_MIB,
		Am79C930_Local_MIB, Fragmentation_Dis, Fill_CFP);
	if(error != 0) return error;

	DP((DN("Tx Buffer: size=%dbyte [%04x-%04x]\n"),
		lm.Tx_Buffer_Size,
		lm.Tx_Buffer_Offset,
		lm.Tx_Buffer_Offset + lm.Tx_Buffer_Size - 1));
	DP((DN("Rx Buffer: size=%dbyte [%04x-%04x]\n"),
		lm.Rx_Buffer_Size,
		lm.Rx_Buffer_Offset,
		lm.Rx_Buffer_Offset + lm.Rx_Buffer_Size - 1));

	sc->ux_rxbufoffset = lm.Rx_Buffer_Offset;
	sc->ux_rxbufsize = lm.Rx_Buffer_Size;

	error = ux_initmibs(sc);
	if(error != 0) return error;

	ux_inittxbuf(sc, lm.Tx_Buffer_Offset, lm.Tx_Buffer_Size);

	/* start INTR */
	ux_setintrmask(sc, 0 /*AM79C930_INTR_CMND*/);
	ux_enableintr(sc, 1);

	ux_inittx(sc);
	ux_initrx(sc);

#if 0
	{
		struct Am79C930_ESS_ID essid;
		bzero(&essid, sizeof(essid));
#if 0
		essid.Type = 0;
		essid.Length = 20;
		bcopy("4D0041TDIRCACKYOTO01", essid.Data, 20);
#endif
		ux_beginscan(sc, &essid);
	}
#endif

	return 0;
}

/*
 *  dealing with 802.11 header
 */

STATIC int
ux_prepend80211header(struct ux_softc *sc, struct mbuf **mp)
{
	struct mbuf *m = *mp;
	struct Am79C930_80211_Header *wh;
	struct ether_header *eh;

	M_PREPEND(m, sizeof(*wh), M_DONTWAIT);
	if(m == 0) {
		*mp = 0;
		return ENOBUFS;
	}

	m = m_pullup(m, sizeof(*wh) + sizeof(*eh));
	if(m == 0) {
		*mp = 0;
		return ENOBUFS;
	}
	wh = mtod(m, struct Am79C930_80211_Header *);
	eh = (struct ether_header *)(wh + 1);

	wh->Frame_Control = AM79C930_80211HDR_FC_PROTOCOL_VERSION
				| AM79C930_80211HDR_FC_TYPE_DATA;
	wh->Duration_ID = 0;
	wh->Sequence_Control = 0;
	bcopy(eh->ether_dhost, wh->Address[0], sizeof(wh->Address[0]));
	bcopy(eh->ether_shost, wh->Address[1], sizeof(wh->Address[1]));
	bcopy(sc->ux_bssid,    wh->Address[2], sizeof(wh->Address[2]));
	
	*mp = m;
	return 0;
}

STATIC void
ux_remove80211header(struct ux_softc *sc, struct mbuf **mp)
{
	m_adj(*mp, sizeof(struct Am79C930_80211_Header));
	return;
}

/*
 *  reset card
 */

STATIC void
ux_resetstart(struct ux_softc *sc)
{
#if 0
	untimeout(ux_resettimo, sc);

	/* reset card */
	UX_IO_WRITE1(sc, AM79C930_IOREG_GCR_CORESET, AM79C930_IOREG_GCR);
	UX_IO_WRITE1(sc, AM79C930_IOREG_GCR_CORESET
			|AM79C930_IOREG_GCR_SWRESET, AM79C930_IOREG_GCR);
	DELAY(10);
	UX_IO_WRITE1(sc, AM79C930_IOREG_GCR_CORESET, AM79C930_IOREG_GCR);
	DELAY(10);

	/* XXX -- clear self test status */
	UX_MEM_WRITE1(sc, 0,
		UX_OFFSETOF(Am79C930_Control_Status_Block, Self_Test_Status,
			AM79C930_ADDR_CONTROL_STATUS_BLOCK));

	/* start firmware */
	UX_IO_WRITE1(sc, 0, AM79C930_IOREG_GCR);

	printf("%s: ux_resetstart().\n", sc->ux_devname);
#endif

	/* wait for finishing self test (about 2 sec) */
	sc->ux_rsttimocnt = 500;
	sc->ux_prevsts = -1;
	ux_resettimo(sc);

	return;
}

STATIC void
ux_resettimo(void *arg)
{
	struct ux_softc *sc = arg;
	char *msg;
	int error, status;

	if(sc->ux_rsttimocnt <= 0) {
		printf("%s: fail to initialize due to timeout. disabled.\n",
					sc->ux_devname);
		return;
	};
	sc->ux_rsttimocnt--;

	error = ux_checkselfteststatus(sc, &status, &msg);

	if(status != sc->ux_prevsts) {
		DP((DN("change Self Test Status(0x%02x): %s\n"), status, msg));
		sc->ux_prevsts = status;
	}

	switch(error) {
	case EBUSY:
		timeout(ux_resettimo, sc, hz / 100);
		return;
	case 0:		/* ok ! */
		break;
	case EIO:	/* NG... */
		printf("%s: fail to initialize. disabled.\n", sc->ux_devname);
		return;
	default:
		printf("%s: bug in driver. :-p\n", sc->ux_devname);
		return;
	}

	/* the card becomes the normal state. */
	error = ux_checkfirmware(sc);
	if(error != 0) {
		printf("%s: fail to initialize. disabled.\n", sc->ux_devname);
		return;
	}
		
	printf("%s: complete initialization.\n", sc->ux_devname);

	return;
}

/*
 *  exported function
 */

/* attach logical interface. don't touch the chip! */
void
ux_attach(struct ux_softc *sc)
{
	struct ifnet *ifp = &sc->ux_if;

	ifp->if_mtu = 1500;	/* XXX */
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_baudrate = 2000000;

	if_attach(ifp);
	return;
}

/* attach the chip and make it available */
void
ux_enable(struct ux_softc *sc)
{
	ux_uint48_t mac;

	if(sc->ux_enabled) return;

	if(ux_checkfirmware(sc) != 0) {
		printf("%s: failed to enable.\n", sc->ux_devname);
		return;
	}

	if(ux_getmacaddr(sc, &mac) != 0) {
		printf("%s: ux_getmacaddr() failed.\n", sc->ux_devname);
		return;
	}

	DP((DN("macaddr=%02x:%02x:%02x:%02x:%02x:%02x\n"),
			mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]));

	if(bcmp(&mac, &sc->ux_macaddr, sizeof(mac)) == 0) {
		printf("%s: the same card returns.\n", sc->ux_devname);
	} else {
		printf("%s: a new card comes.\n", sc->ux_devname);
	}

	bcopy(&mac, &sc->ux_macaddr, ETHER_ADDR_LEN);

	sc->ux_if.if_snd.ifq_maxlen = ifqmaxlen;

	ux_inittxrx(sc);
#if 1
	{
		struct Am79C930_ESS_ID essid;
		bzero(&essid, sizeof(essid));
		ux_beginscan(sc, &essid);
	}
#endif

	sc->ux_enabled = 1;
	return;
}

void
ux_disable(struct ux_softc *sc)
{
	if(! sc->ux_enabled) return;

	sc->ux_enabled = 0;
	return;
}

int
ux_intr(struct ux_softc *sc)
{
	u_int16_t status, done;

	status = ux_beginintr(sc);

	if(status &	~(AM79C930_INTR_CMND
			| AM79C930_INTR_SCAN_CMPLT
			| AM79C930_INTR_TX
			| AM79C930_INTR_RX
			| AM79C930_INTR_RX_DATA)) {
		DP9((DN("ux_intr(): status=0x%04x\n"), status));
	}

	if(status & AM79C930_INTR_TX) {
		ux_cleanuptx(sc);
	}
	if(status & AM79C930_INTR_RX_PS_POLL_MGMT) {
		ux_scanrxqueue(sc, &sc->ux_rxdeschead_mgmt);
	}
	if(status & AM79C930_INTR_RX_DATA) {
		ux_scanrxqueue(sc, &sc->ux_rxdeschead);
	}
	if(status & AM79C930_INTR_SCAN_CMPLT) {
		ux_endscan(sc);
	}

	return 1;
}

void
ux_start(struct ux_softc *sc)
{
	struct ifnet *ifp = &sc->ux_if;
	struct mbuf *m;
	int error;
	int count;

	for(count = 0; count < 30; count++) {
		IF_DEQUEUE(&ifp->if_snd, m);
		if(m == 0) break;

		DP9((DN("ux_start(): count=%d\n"), count));

		if(! sc->ux_rawmacbpf) ux_tap(sc, m);

		error = ux_prepend80211header(sc, &m);
		if(error != 0) {
			IF_PREPEND(&ifp->if_snd, m);
			DP((DN("ux_start(): fail to append 802.11 header."
						" error=%d\n"), error));
			ifp->if_oerrors++;
			break;
		}

		if(sc->ux_rawmacbpf) ux_tap(sc, m);

		error = ux_appendtxqueue(sc, &sc->ux_txdesctail, m);
		if(error != 0) {
			ux_remove80211header(sc, &m);
			IF_PREPEND(&ifp->if_snd, m);
			if(error == ENOBUFS) {
				DP((DN("ux_start(): return pkt to queue.\n")));
			} else {
				ifp->if_oerrors++;
				DP((DN("ux_start(): fail to append."
						" error=%d\n"), error));
			}
			break;
		}
		ifp->if_opackets++;

		m_freem(m);
	}
#if 0
	if(count > 0) ifp->if_flags |= IFF_OACTIVE;
#endif
	return;
}

void
ux_reset(struct ux_softc *sc)
{
	DP((DN("ux_reset()\n")));
	ux_resetstart(sc);
}

int
ux_init(struct ux_softc *sc)
{
	int s = splimp();

	DP((DN("ux_init()\n")));

	/* let's begin communication. */

	ux_inittxrx(sc);

	sc->ux_if.if_flags &= ~IFF_OACTIVE;
	sc->ux_if.if_flags |= IFF_RUNNING;

	ux_start(sc);

	splx(s);
	
	return 0;
}

int
ux_ioctl(struct ux_softc *sc, u_long cmd, caddr_t data)
{
	struct ifnet *ifp = &sc->ux_if;
	struct ifaddr *ifa = (struct ifaddr *)data;
	int error = EINVAL;
	int s = splimp();
		
	switch(cmd) {
	case SIOCGIFADDR:
	case SIOCSIFADDR:
		error = ether_ioctl(ifp, cmd, data);
		break;
	case SIOCSIFFLAGS:
		if(ifp->if_flags & IFF_UP) {
			if((ifp->if_flags & IFF_RUNNING) == 0) ux_init(sc);
		} else {
			ifp->if_flags &= ~IFF_RUNNING;
		}
		
		error = ux_setpromiscuous(sc, ifp->if_flags & IFF_PROMISC);
		break;

	case UXIOCEXECCMD: {
		struct uxioctl_execcmd *ec = (void *)data;
		struct Am79C930_Command_Block *cb = &ec->cb;

		if(cb->Command == 0) {
			error = EINVAL;
			break;
		}

		UX_MEM_WRITEM(sc,
			&cb->Command_Parameters, sizeof(cb->Command_Parameters),
			UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
					AM79C930_ADDR_COMMAND_BLOCK));

		error = ux_execcommand(sc, cb->Command, &cb->Command_Status);
		
		cb->Error_Offset = UX_MEM_READ1(sc,
			UX_OFFSETOF(Am79C930_Command_Block, Error_Offset,
					AM79C930_ADDR_COMMAND_BLOCK));
		UX_MEM_READM(sc,
			&cb->Command_Parameters, sizeof(cb->Command_Parameters),
			UX_OFFSETOF(Am79C930_Command_Block, Command_Parameters,
					AM79C930_ADDR_COMMAND_BLOCK));
		} break;

	case UXIOCREADMEM: {
		struct uxioctl_cardmem *cm = (void *)data;

		if(cm->size > sizeof(cm->data)) cm->size = sizeof(cm->data);
		UX_MEM_READM(sc, cm->data, cm->size, cm->addr);
		error = 0;

		} break;
		
	case UXIOCSRAWMACBPF: {
		struct uxioctl_int *ui = (void *)data;
		sc->ux_rawmacbpf = ui->i;
		error = 0;
		} break;

	case UXIOCSESSID: {
		struct uxioctl_essid *ue = (void *)data;
		struct Am79C930_ESS_ID essid;
		
		if(ue->length > sizeof(essid.Data)) {
			error = EINVAL;
			break;
		}
		essid.Type = 0;
		essid.Length = ue->length;
		bcopy(&ue->essid, &essid.Data, ue->length);

		ux_beginscan(sc, &essid);

		error = 0;
		
		} break;
	case UXIOCGESSID:

	default:
		error = EINVAL;
		break;
	}

	splx(s);
	return error;
}

/* EOF */
